"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

function _micromatch() {
  const data = require("micromatch");

  _micromatch = function () {
    return data;
  };

  return data;
}

function _path() {
  const data = require("path");

  _path = function () {
    return data;
  };

  return data;
}

var _loadParcelPlugin = _interopRequireDefault(require("./loadParcelPlugin"));

var _projectPath = require("./projectPath");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class ParcelConfig {
  constructor(config, options) {
    this.options = options;
    this.filePath = config.filePath;
    this.resolvers = config.resolvers || [];
    this.transformers = config.transformers || {};
    this.runtimes = config.runtimes || [];
    this.bundler = config.bundler;
    this.namers = config.namers || [];
    this.packagers = config.packagers || {};
    this.optimizers = config.optimizers || {};
    this.reporters = config.reporters || [];
    this.validators = config.validators || {};
    this.pluginCache = new Map();
    this.regexCache = new Map();
  }

  static deserialize(serialized) {
    return new ParcelConfig(serialized.config, serialized.options);
  }

  getConfig() {
    return {
      filePath: this.filePath,
      resolvers: this.resolvers,
      transformers: this.transformers,
      validators: this.validators,
      runtimes: this.runtimes,
      bundler: this.bundler,
      namers: this.namers,
      packagers: this.packagers,
      optimizers: this.optimizers,
      reporters: this.reporters
    };
  }

  serialize() {
    return {
      $$raw: false,
      config: this.getConfig(),
      options: this.options
    };
  }

  _loadPlugin(node) {
    let plugin = this.pluginCache.get(node.packageName);

    if (plugin) {
      return plugin;
    }

    plugin = (0, _loadParcelPlugin.default)(node.packageName, (0, _projectPath.fromProjectPath)(this.options.projectRoot, node.resolveFrom), node.keyPath, this.options);
    this.pluginCache.set(node.packageName, plugin);
    return plugin;
  }

  async loadPlugin(node) {
    let plugin = await this._loadPlugin(node);
    return { ...plugin,
      name: node.packageName,
      keyPath: node.keyPath
    };
  }

  invalidatePlugin(packageName) {
    this.pluginCache.delete(packageName);
  }

  loadPlugins(plugins) {
    return Promise.all(plugins.map(p => this.loadPlugin(p)));
  }

  _getResolverNodes() {
    if (this.resolvers.length === 0) {
      throw new Error('No resolver plugins specified in .parcelrc config');
    }

    return this.resolvers;
  }

  getResolverNames() {
    return this._getResolverNodes().map(r => r.packageName);
  }

  getResolvers() {
    return this.loadPlugins(this._getResolverNodes());
  }

  _getValidatorNodes(filePath) {
    let validators = this.matchGlobMapPipelines(filePath, this.validators) || [];
    return validators;
  }

  getValidatorNames(filePath) {
    let validators = this._getValidatorNodes(filePath);

    return validators.map(v => v.packageName);
  }

  getValidators(filePath) {
    let validators = this._getValidatorNodes(filePath);

    return this.loadPlugins(validators);
  }

  getNamedPipelines() {
    return Object.keys(this.transformers).filter(glob => glob.includes(':')).map(glob => glob.split(':')[0]);
  }

  _getTransformerNodes(filePath, pipeline, allowEmpty) {
    let transformers = this.matchGlobMapPipelines(filePath, this.transformers, pipeline);

    if (!transformers || transformers.length === 0) {
      if (allowEmpty) {
        return [];
      }

      throw new Error(`No transformers found for ${(0, _projectPath.fromProjectPathRelative)(filePath)}` + (pipeline != null ? ` with pipeline: '${pipeline}'` : '') + '.');
    }

    return transformers;
  }

  getTransformerNames(filePath, pipeline, allowEmpty) {
    let transformers = this._getTransformerNodes(filePath, pipeline, allowEmpty);

    return transformers.map(t => t.packageName);
  }

  getTransformers(filePath, pipeline, allowEmpty) {
    return this.loadPlugins(this._getTransformerNodes(filePath, pipeline, allowEmpty));
  }

  getBundlerName() {
    if (!this.bundler) {
      throw new Error('No bundler specified in .parcelrc config');
    }

    return this.bundler.packageName;
  }

  getBundler() {
    if (!this.bundler) {
      throw new Error('No bundler specified in .parcelrc config');
    }

    return this.loadPlugin(this.bundler);
  }

  getNamers() {
    if (this.namers.length === 0) {
      throw new Error('No namer plugins specified in .parcelrc config');
    }

    return this.loadPlugins(this.namers);
  }

  getRuntimes() {
    if (!this.runtimes) {
      return Promise.resolve([]);
    }

    return this.loadPlugins(this.runtimes);
  }

  _getPackagerNode(filePath) {
    let packagerName = this.matchGlobMap((0, _projectPath.toProjectPathUnsafe)(filePath), this.packagers);

    if (!packagerName) {
      throw new Error(`No packager found for "${filePath}".`);
    }

    return packagerName;
  }

  getPackagerName(filePath) {
    return this._getPackagerNode(filePath).packageName;
  }

  getPackager(filePath) {
    let packager = this._getPackagerNode(filePath);

    return this.loadPlugin(packager);
  }

  _getOptimizerNodes(filePath, pipeline) {
    var _this$matchGlobMapPip;

    // If a pipeline is specified, but it doesn't exist in the optimizers config, ignore it.
    // Pipelines for bundles come from their entry assets, so the pipeline likely exists in transformers.
    if (pipeline) {
      let prefix = pipeline + ':';

      if (!Object.keys(this.optimizers).some(glob => glob.startsWith(prefix))) {
        pipeline = null;
      }
    }

    return (_this$matchGlobMapPip = this.matchGlobMapPipelines((0, _projectPath.toProjectPathUnsafe)(filePath), this.optimizers, pipeline)) !== null && _this$matchGlobMapPip !== void 0 ? _this$matchGlobMapPip : [];
  }

  getOptimizerNames(filePath, pipeline) {
    let optimizers = this._getOptimizerNodes(filePath, pipeline);

    return optimizers.map(o => o.packageName);
  }

  getOptimizers(filePath, pipeline) {
    let optimizers = this._getOptimizerNodes(filePath, pipeline);

    if (optimizers.length === 0) {
      return Promise.resolve([]);
    }

    return this.loadPlugins(optimizers);
  }

  getReporters() {
    return this.loadPlugins(this.reporters);
  }

  isGlobMatch(projectPath, pattern, pipeline) {
    // glob's shouldn't be dependant on absolute paths anyway
    let filePath = (0, _projectPath.fromProjectPathRelative)(projectPath);
    let [patternPipeline, patternGlob] = pattern.split(':');

    if (!patternGlob) {
      patternGlob = patternPipeline;
      patternPipeline = null;
    }

    let re = this.regexCache.get(patternGlob);

    if (!re) {
      re = (0, _micromatch().makeRe)(patternGlob, {
        dot: true,
        nocase: true
      });
      this.regexCache.set(patternGlob, re);
    }

    return (pipeline === patternPipeline || !pipeline && !patternPipeline) && (re.test(filePath) || re.test((0, _path().basename)(filePath)));
  }

  matchGlobMap(filePath, globMap) {
    for (let pattern in globMap) {
      if (this.isGlobMatch(filePath, pattern)) {
        return globMap[pattern];
      }
    }

    return null;
  }

  matchGlobMapPipelines(filePath, globMap, pipeline) {
    let matches = [];

    if (pipeline) {
      // If a pipeline is requested, a the glob needs to match exactly
      let exactMatch;

      for (let pattern in globMap) {
        if (this.isGlobMatch(filePath, pattern, pipeline)) {
          exactMatch = globMap[pattern];
          break;
        }
      }

      if (!exactMatch) {
        return [];
      } else {
        matches.push(exactMatch);
      }
    }

    for (let pattern in globMap) {
      if (this.isGlobMatch(filePath, pattern)) {
        matches.push(globMap[pattern]);
      }
    }

    let flatten = () => {
      let pipeline = matches.shift() || [];
      let spreadIndex = pipeline.indexOf('...');

      if (spreadIndex >= 0) {
        pipeline = [...pipeline.slice(0, spreadIndex), ...flatten(), ...pipeline.slice(spreadIndex + 1)];
      }

      if (pipeline.includes('...')) {
        throw new Error('Only one spread parameter can be included in a config pipeline');
      }

      return pipeline;
    };

    let res = flatten(); // $FlowFixMe afaik this should work

    return res;
  }

}

exports.default = ParcelConfig;